#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
py2c.py

train_mnist.pyから学習したパラメータを
C合成用ヘッダファイルとして出力

Developed by H. Nakahara
             Tokyo Institute of Technology

使用方法, train_mnist.py で学習後、ファイルが出力されるので

$python py2c.py -m model.bin.cnn.npz 

と入力すると, Cシミュ用のヘッダ(param_cnn_float.h)と
AOCL用のヘッダ(param_cnn_aocl.h)が出力されます.

"""

import argparse
import time
import numpy as np
import math
from chainer import serializers

# load parameters -----------------------------------
parser = argparse.ArgumentParser(description='Converter from Pre-trained model:')
parser.add_argument('--model', '-m', default='',
                    help='Load a pre-trained model') # --model (学習済みCNN.npz) を読込み
parser.add_argument('--int', '-i', default=8, type=int,
                    help='bit precision for integer representation layer') # 固定小数点表現レイヤーのビット精度(デフォルト8ビット)
args = parser.parse_args()

# Load
if args.model:
    print('Load model from %s' % args.model)
    dat = np.load(args.model)
else:
    print('ERROR: Specify pr-trained model (*.npz)')
    exit()

print("Bit precision = %d" % args.int)
precision = args.int

# 学習したパラメータを表示
for key, value in dat.items():
    print ("key=%s" % (key)), # キーワードを表示
    print (",dimension="),
    print (value.shape) # 次元を表示

# train_mnist.py で学習したパラメータを読み出す
param = ''
param += "//bit precision = %d\n" % precision
param += "#define NBIT %d\n" % precision

import re

headerfile_float = ''
headerfile_aocl  = ''

declaration = ''

for key, value in dat.items():
    res = key.split('/')
    print(res)
    print(res[1])

    if 'conv' in res[1] and 'W' in res[2]:
        print('find weight for convolutional layer in %s' % key)

        ksiz   = value.shape[2]
        n_ofeat = value.shape[0]
        n_infeat = value.shape[1]
        param_name = res[1] + res[2]
    
        # floating point
        param_float = "const float %s[%d][%d][%d][%d] = {" % (param_name,n_ofeat,n_infeat,ksiz,ksiz)
        param_aocl  = "constant float %s[%d][%d][%d][%d] = {" % (param_name,n_ofeat,n_infeat,ksiz,ksiz)

        dat = ''
        for ofeat in range(n_ofeat):
            if ofeat == 0:
                dat += "{"
            else:
                dat += ",{"            

            for infeat in range(n_infeat):
                if infeat == 0:
                    dat += "{"
                else:
                    dat += ",{"            

                for ky in range(ksiz):
                    if ky == 0:
                        dat += "{"
                    else:
                        dat += ",{"            

                    for kx in range(ksiz):
                        tmp = value[ofeat,infeat,ky,kx]
                        if kx == 0:
                            dat += "%f" % tmp
                        else:
                            dat += ",%f" % tmp       

                    dat += "}"
                dat += "}"
            dat += "}"
        dat += "};\n"
#        print(param_float)


        headerfile_float += (param_float + dat)
        headerfile_aocl  += (param_aocl + dat)

    else:
        print('find other %s' % key)
        # if # dims = 0, then, just skip
        if value.ndim == 0:
            pass
        elif value.ndim == 1:
            n_ofeat = value.shape[0]
            param_name = res[1] + res[2]
        
            # floating point 
            param_float = "const float %s[%d] = {" % (param_name,n_ofeat)

            param_aocl  = "constant float %s[%d] = {" % (param_name,n_ofeat)

            dat = ''
            for ofeat in range(n_ofeat):
                tmp = value[ofeat]
                if ofeat < n_ofeat - 1:
                    dat += "%f," % tmp
                else:
                    dat += "%f" % tmp
            dat += "};\n"
#            print(param_float)

            headerfile_float += (param_float + dat)
            headerfile_aocl += (param_aocl + dat)


        elif value.ndim == 2:

            n_ofeat = value.shape[0]
            n_infeat = value.shape[1]
            param_name = res[1] + res[2]

            param_float = "const float %s[%d][%d] = {" % (param_name,n_ofeat,n_infeat)
            param_aocl = "constant float %s[%d][%d] = {" % (param_name,n_ofeat,n_infeat)

            dat = ''
            for ofeat in range(n_ofeat):
                if ofeat == 0:
                    dat += "{"
                else:
                    dat += ",{"            

                for infeat in range(n_infeat):
                    tmp = value[ofeat,infeat]

                    if infeat < n_infeat - 1:
                        dat += "%f," % tmp
                    else:
                        dat += "%f" % tmp
                dat += "}"
            dat += "};\n"

            headerfile_float += (param_float+dat)
            headerfile_aocl += (param_aocl+dat)

        else:
            print("ERROR: UNDEFINED PARAMETERS %s EXIST" % key)
            exit()

# Output C/C++ headerfile
# floating point representation for C/C++
fp = open('param_cnn_float.h', 'w')
print('Generate param_cnn_float.h')
fp.write(headerfile_float)
fp.close()

# constant representation for AOCL
fp = open('param_cnn_aocl.h', 'w')
print('Generate param_cnn_aocl.h')
fp.write(headerfile_aocl)
fp.close()

# -------------------------------------------------------------------------------
# END OF PROGRAM
# -------------------------------------------------------------------------------
